# -*- coding: ISO-8859-1 -*-

import operator, types, persistence
from objs import *

#-----------------------------------------------------------------------------

# La tabla de smbolos

class RuntimeSymbolTable(object):
	def __init__(self):
		self.blockVars = {}
		self.blockStack = []
		self.vars = {}
		self.varStack = []
		self.currentClass = None
		self.classStack = []
		self.currentInstance = None
		self.instanceStack = []

	def getCurrentClass(self): return self.currentClass
	def getCurrentInstance(self): return self.currentInstance

	def methodEnter(self, instance, method, args):
		self.classStack.append(self.currentClass)
		self.currentClass = method.getClass()

		self.blockStack.append(self.blockVars)
		self.blockVars = {}

		self.varStack.append(self.vars)
		self.vars = {}

		self.instanceStack.append(self.currentInstance)
		self.currentInstance = instance

		# Metemos las variables miembro de la instancia
		for name, ref in instance.getFields(self.currentClass).items():
			self.setVar(name, ref)
		# Ahora los argumentos que le pasamos al mtodo
		for (name, klass), ref in zip(method.getArgs(), args):
			self.setVar(name, JRef(klass, ref.getInstance()))
		# Y por ltimo el this
		self.setVar('this', JRef(self.currentClass, instance))

	def methodExit(self):
		self.vars = self.varStack.pop()
		self.blockVars = self.blockStack.pop()
		self.currentClass = self.classStack.pop()
		self.currentInstance = self.instanceStack.pop()

	def blockEnter(self):
		self.blockStack.append(self.blockVars)
		self.blockVars = {}
		self.varStack.append(self.vars)
		self.vars = self.vars.copy()

	def blockExit(self):
		self.blockVars = self.blockStack.pop()
		self.vars = self.varStack.pop()

	def setVar(self, name, ref):
		self.blockVars[name] = ref
		self.vars[name] = ref

	def getVar(self, name):
		#return self.vars[name]
		result=self.vars.get(name)
		if result:
			return result
		for frame in self.varStack:
			if frame.get(name):
				return frame[name]
		return None

		

#-----------------------------------------------------------------------------

# El intrprete

class Interpreter(object):
	def __init__(self, application, classes = {}):
		self.application = application
		self.symbolTable = RuntimeSymbolTable()
		self.classes = classes
		self.manager = persistence.Manager(self, application)
		from threading import Semaphore
		self.semaphore=Semaphore(0)

	def acquire(self):
		self.semaphore.acquire()
		
	def terminate(self):
		self.manager.terminate()
		self.semaphore.release()

	def getSymbolTable(self): return self.symbolTable
	def getClassByName(self, name): return self.classes[name]
	def getPersistenceManager(self): return self.manager
	def getApplication(self): return self.application

	def parse(self, l):
		if l and operator.isSequenceType(l) \
			and not isinstance(l, types.StringTypes):
			f = getattr(self, 'visit_' + l[0])
			return apply(f, l[1:])
		else:
			return l

	def visit_main(self, block):
		self.parse(block)

	def visit_block(self, *statements):
		for s in statements: self.parse(s)

	def visit_decl(self, klass, *vars):
		for name, init in vars:
			ref = self.parse(init)
			inst = ref.getInstance()
			self.symbolTable.setVar(name, JRef(klass, inst))

	def visit_return(self, exprs):
		raise JReturn(self.parse(exprs).getInstance())

	def visit_voidreturn(self):
		raise JReturn(voidInstance)

	def visit_exprs(self, *exprs):
		e = [self.parse(expr) for expr in exprs]
		return e[-1]

	def visit_assign(self, lval, rval):
		l = self.parse(lval)
		r = self.parse(rval)
		l.setInstanceFrom(r)
		return l

	def visit_is(self, lval, rval):
		l = self.parse(lval).getInstance()
		r = self.parse(rval).getInstance()
		return JRef(boolClass, makeBool(l is r))

	def visit_land(self, lval, rval):
		l = self.parse(lval)
		if l.isNull(): result = False
		else: result = not self.parse(rval).isNull()
		return JRef(boolClass, makeBool(result))

	def visit_lor(self, lval, rval):
		l = self.parse(lval)
		if l.isNull(): result = not self.parse(rval).isNull()
		else: result = True
		return JRef(boolClass, makeBool(result))

	def visit_lnot(self, val):
		v = self.parse(val)
		return JRef(boolClass, makeBool(v.isNull()))

	def visit_new(self, klass, ctor, ctor_args):
		inst = klass.newInstance()
		a = [self.parse(arg) for arg in ctor_args]
		ctor.invoke(self, inst, a)
		r = JRef(klass, inst)
		return r

	def visit_null(self):
		return JRef(nullClass, nullInstance)

	def visit_this(self):
		return JRef(self.symbolTable.getCurrentClass(),
			self.symbolTable.getCurrentInstance())

	def visit_true(self):
		return JRef(boolClass, boolInstance)

	def visit_ident(self, ident):
		r = self.symbolTable.getVar(ident)
		return self.symbolTable.getVar(ident)

	def visit_int(self, i):
		ii = integerClass.newInstance()
		ii.setValue(i)
		return JRef(integerClass, ii)

	def visit_str(self, s):
		si = stringClass.newInstance()
		si.setValue(s)
		return JRef(stringClass, si)

	def visit_invokevirtual(self, ref, mangledName, args):
		r = self.parse(ref)
		if r.isNull():
			raise 'La instancia es nula'
		inst = r.getInstance()
		klass = inst.getClass()
		method = klass.getMethodByMangledName(mangledName)
		a = [self.parse(arg) for arg in args]
		returnType = method.getReturnType()
		result = method.invoke(self, inst, a)
		#if not result: result = nullInstance
		return JRef(returnType, result)

	def visit_invokespecial(self, ref, method, args):
		r = self.parse(ref)
		if r.isNull():
			raise 'La referencia es nula'
		inst = r.getInstance()
		a = [self.parse(arg) for arg in args]
		returnType = method.getReturnType()
		result = method.invoke(self, inst, a)
		#if not result: result = nullInstance
		return JRef(returnType, result)

	def visit_field(self, ref, klass, name):
		r = self.parse(ref)
		if r.isNull(): raise 'La referencia es nula'
		inst = r.getInstance()
		return inst.getField(klass, name)

	def visit_if(self, cond, ifTrue, ifFalse):
		c = self.parse(cond)
		if c.isNull(): return self.parse(ifFalse)
		else: return self.parse(ifTrue)

	def visit_while(self, cond, code):
		while True:
			c = self.parse(cond)
			if c.isNull(): break
			self.parse(code)

	def visit_for(self, init, cond, step, code):
		# Como en un for se pueden declarar variables locales en su
		# seccin de inicializacin, tengo que hacer un blockEnter
		self.symbolTable.blockEnter()
		self.parse(init)
		while True:
			c = self.parse(cond)
			if c.isNull(): break
			self.parse(code)
			self.parse(step)
		self.symbolTable.blockExit()

	def visit_reify(self, reify_code):
		reify_code.execute()

	def visit_upcast(self, klass, ref):
		r = self.parse(ref)
		return JRef(klass, r.getInstance())

	def visit_downcast(self, klass, ref):
		r = self.parse(ref)
		inst = r.getInstance()
		if inst.isA(klass):
			return JRef(klass, inst)
		else:
			raise 'La instancia no es un %s' % klass
