"Language Specification Module"

import string
import metaLang
import appSpec

# ************************** Language *****************************

class Language:
    "Language Definition Class; defines each application's language"
    
    def __init__(self,name):
        "Language Constructor"
        # Language name
        self.name=name
        # Lexical Specification: Dictionary of NonTerminal objets that recognises tokens
        self.lexicalSpec={}
        # Syntactic Specification: Dictionary of NonTerminal objects
        self.syntacticSpec={}
        # List of tokens to skip
        self.skipSpec=[]
        # List of tokens not to skip
        self.notSkipSpec=[]
        # Language NonTerminal Start Symbol (don't know)
        self.start=0

    def describe(self):
        "Gives us information related to the Language at runtime"
        # Language description
        description="Language: "+self.name+".\n\n"
        # Scanner-Specification Rules
        description=description+"Scanner-Specification Rules:\n"
        keys=self.lexicalSpec.keys()
        keys.sort()
        for key in keys:
            rule=self.lexicalSpec[key]
            description=description+"\tRule description: "+rule.description+"\n"
            description=description+"\t"+rule.name+" -> "
            firstRule=1
            for option in rule.options:
                if firstRule:
                    firstRule=0
                else:
                    description=description+"\t\t| "
                for node in option.nodes:
                    if node.__class__.__name__=="NonTerminal":
                        description=description+node.name+" "
                    elif node.__class__.__name__=="Terminal":
                        description=description+"\""+node.value+"\" "
                    elif node.__class__.__name__=="Lambda":
                        description=description+"Lambda "
                    else:
                        raise "Unknown instance"
                for action in option.actions:
                    description=description+"{"+action.action+"}\n"
                description=description+"\n"
            description=description+"\n"
        # Parser-Specification Rules
        description=description+"Parser-Specification Rules.\n"
        description=description+"Syntactic Start Symbol: "+self.start.name+"\n\n"
        keys=self.syntacticSpec.keys()
        keys.sort()
        for key in keys:
            rule=self.syntacticSpec[key]
            description=description+"\tRule description: "+rule.description+"\n"
            description=description+"\t"+rule.name+" -> "
            firstRule=1
            for option in rule.options:
                if firstRule:
                    firstRule=0
                else:
                    description=description+"\t\t| "
                for node in option.nodes:
                    if node.__class__.__name__=="NonTerminal":
                        description=description+node.name+" "
                    elif node.__class__.__name__=="Lambda":
                        description=description+"Lambda "
                    else:
                        raise "Unknown instance"
                for action in option.actions:
                    description=description+"{"+action.action+"}"
                description=description+"\n"
            description=description+"\n"
        # Skip tokens
        description=description+"Skip Tokens:\n"
        for token in self.skipSpec:
            description=description+"\t"+token.value+"\n"
        description=description+"\n"
        # NotSkip tokens
        description=description+"NotSkip Tokens:\n"
        for token in self.notSkipSpec:
            description=description+"\t"+token.value+"\n"
        description=description+"\n"
        return description
    
# ************************** NonTerminal ******************************

class NonTerminal:
    "NonTerminal symbols, either in the right or in left part of a free-context grammar rule"
    
    def __init__(self,name,description=""):
        "NonTerminal Constructor"
        # Grammar symbol identification
        self.name=name
        # Right Part of the rule
        # If the NT has no description, it's on a right part of a rule
        if description=="":
            self.options=0 # No right part
        else:
            self.options=[Right()] # At first, one right part
        # Rule description
        self.description=description

    def getName(self):
        "Returns its name"
        return self.name
    
    def describe(self):
        "Describes the grammatical element and its right part"
        # Located on the left part of the rule
        if self.options!=0:
            # Left Part
            description="NonTerminal Grammar Symbol \""+self.name+"\" on the left part of a rule "
            description=description+"described as \""+self.description+"\". "
            description=description+"Rule definition:\n"
            description=description+"\t"+self.name+" -> "
            firstRule=1
            for option in self.options:
                if firstRule:
                    firstRule=0
                else:
                    description=description+"\t\t| "
                for node in option.nodes:
                    if node.__class__.__name__=="Terminal":
                        description=description+"\""+node.value+"\" "
                    elif node.__class__.__name__=="NonTerminal":
                        description=description+node.name+" "
                    elif node.__class__.__name__=="Lambda":
                        description=description+"Lambda "
                    else:
                        raise "Unknown instance"
                for action in option.actions:
                    description=description+"{"+action.action+"}"
                description=description+"\n"
        else:
            # Right Part
            description="NonTerminal Grammar Symbol \""+self.name+"\" on the right part of a rule."
        return description

    def recognise(self,gi):
        "Finds out whether the grammatical symbol recognises the input. Returns its application tree."
        if self.options==0:
            # Located on the right part; must find its rule
            if gi.application.language.syntacticSpec.has_key(self.name):
                rule=gi.application.language.syntacticSpec[self.name]
            elif gi.application.language.lexicalSpec.has_key(self.name):
                rule=gi.application.language.lexicalSpec[self.name]
            else:
                raise "The NonTerminal "+self.name+" doesn't have a rule defined."
        else:
            rule=self
        # Analises the rule's right-parts
        recognised=0
        initialPosition=gi.buffer.position
        initialText=gi.buffer.text
        optionNumber=0
        syntacticRule=gi.application.language.syntacticSpec.has_key(rule.name)
        while not(recognised) and optionNumber<len(rule.options):
            # In scanning, skip and notSkip tokens are not cleared (last param)
            """if syntacticRule: # NoComment=TraceParsing
                ruleName="\t: "+rule.getName()+" ->" # NoComment=TraceParsing 
                for node in rule.options[optionNumber].nodes: # NoComment=TraceParsing
                    ruleName=ruleName+" "+node.getName() # NoComment=TraceParsing
                print ruleName # NoComment=TraceParsing"""
            recognised=rule.options[optionNumber].recognise(gi,syntacticRule)
            if not(recognised):
                gi.buffer.position=initialPosition
            optionNumber=optionNumber+1
        #print rule.name,' ',recognised # NoComment=TraceParsing
        if recognised:
                #print "* "+gi.buffer.text[len(initialText):] # NoComment=TraceParsing
                treeNode=appSpec.TreeNode(gi.application,self,recognised,rule.options[optionNumber-1],gi.buffer.text[len(initialText):])
        else:
            treeNode=0
        return treeNode

# ************************** Terminal ******************************

class Terminal:
    "Terminal symbols defined as its character sequence (attribute value)"

    def __init__(self,value):
        "Terminal Constructor"
        # Grammar symbol identification
        self.value=value

    def getName(self):
        "Returns its value"
        return self.value
    
    def recognise(self,gi):
        "Finds out whether the grammatical symbol recognises the input"
        # Text to be recognised
        position=gi.buffer.position
        text=gi.buffer.code[position:position+len(self.value)]
        recognised=text==self.value
        treeNode=0
        if not(recognised):
            # Scape characters
            if len(self.value)==2 and self.value[0]=="\\":
                text=gi.buffer.code[position:position+1]
                recognised=(self.value=="\\t" and text=="\t") or (self.value=="\\n" and text=="\n") or (self.value=="\\\\" and text=="\\")                
        if recognised:
            gi.buffer.position=position+len(text)            
            gi.buffer.text=gi.buffer.text+text
            treeNode=appSpec.TreeNode(gi.application,self,[],0,text)
        return treeNode

    def describe(self):
        "Describes the grammatical Terminal symbol"
        return "Terminal Grammar Symbol with value \""+self.value+"\"."
       
# ************************** Reify Grammar Symbol **************************

class Reify(NonTerminal):
    "Reify Symbol exits in every language"

    def __init__(self):
        "Reify rule constructor"
        NonTerminal.__init__(self,"Reify","Reify rule")
        # Initaly no options (no code to be reified)
        self.options=[]
        
    def getName(self):
        "Returns its name"
        return "_REIFY_"
    
    def recognise(self,gi):
        "Finds out if a reify statement is written"
        initialPosition=gi.buffer.position
        initialText=gi.buffer.text
        treeNode=0
        gi.clearBuffer()
        text="REIFY"
        if string.upper(gi.buffer.code[gi.buffer.position:gi.buffer.position+len(text)])==text:
            # Reify Statement
            gi.buffer.position=gi.buffer.position+len(text)
            gi.clearBuffer()
            scanner=metaLang.Scanner(gi.buffer.code[gi.buffer.position:])
            if scanner.getNextToken()!=metaLang.Scanner.CODE:
                raise "Reify instruction must be followed by Python Code between <# and #>."
            else:
                # Each option has its semantic rule to be executed
                option=Right()
                self.options.append(option)
                option.actions=[ReifySemanticAction(scanner.text,gi.nitrO)]
                treeNode=appSpec.TreeNode(gi.application,self,[],option,scanner.text)
                # Update code position
                i=gi.buffer.position+2
                codeLevel=1
                while codeLevel:
                    if gi.buffer.code[i:i+2]=="#>":
                        codeLevel=codeLevel-1
                        i=i+2
                    elif gi.buffer.code[i:i+2]=="<#":
                        codeLevel=codeLevel+1
                        i=i+2
                    else:
                        i=i+1
                gi.buffer.position=i
        else:
            gi.buffer.position=initialPosition
            gi.buffer.text=initialText
        return treeNode
                
    def describe(self):
        "Describes the grammatical Reify symbol"
        return "Terminal Reify Symbol."

# ************************** Lambda Symbol **************************

class Lambda:
    "Lambda symbol that regonises every input"

    def __init__(self):
        # Value as a Terminal Symbol
        self.value="LAMBDA"
    
    def getName(self):
        "Returns its name"
        return "LAMBDA"
    
    def recognise(self,gi):
        "Finds out whether the grammatical symbol recognises the input"
        # Lambda is always true (its text is "")
        return appSpec.TreeNode(gi.application,self,[],0,"")
    
    def describe(self):
        "Describes the grammatical symbol"
        return "Lambda Grammar Symbol."

        
# ************************** Right *********************************

class Right:
    "An option of a free-context grammar rule right-part"

    def __init__(self):
        "Rule Right-Part Constructor"
        # List of grammatical symbols on the right part of the rule
        self.nodes=[]
        # List of semantic rules (actions)
        self.actions=[]

    def recognise(self,gi,clear):
        "Finds out whether the right part recognises the input"
        initialPosition=gi.buffer.position
        #initialText=gi.buffer.text
        recognised=1
        nodeNumber=0
        children=[]
        initialText=gi.buffer.text
        while recognised and nodeNumber<len(self.nodes):
            notSkipText=""
            if clear: 
                notSkipText=gi.clearBuffer()
            recognised=self.nodes[nodeNumber].recognise(gi)
            if recognised:
                # Any notSkip tokens cleared?
                if notSkipText!="":
                    recognised.text=notSkipText+recognised.text
                children.append(recognised)
            nodeNumber=nodeNumber+1
        # Not recognised? Set the buffer to the initial position
        if not(recognised):
            gi.buffer.position=initialPosition
            gi.buffer.text=initialText
            children=0
        #else: # NoComment=TraceParsing
        #    for i in self.nodes:
        #        if i.__class__.__name__=="Terminal" or i.__class__.__name__=="Lambda":
        #            print i.value,' ',
        #        else:
        #            print i.name,' ',
        #    print ' '
        return children

# ************************ Semantic Action ***************************

# *** ADDED FOR PFC
class SemanticActionException:
    def __init__(self, ex0, msg0):
        self.ex0 = ex0
        self.msgs = [msg0]
    def append(self, msg):
        self.msgs.append(msg)
    def __str__(self):
        msgs = self.msgs[:]
        msgs.reverse()
        return "Exception in semantic action; trace (current last):\n" \
            + ('*'*10) + "\n" + (('-'*10) + "\n").join(msgs) + ('*'*10) \
            + "\nOriginal exception: " + str(self.ex0)
# *** END ADDED FOR PFC

class SemanticAction:
    "A semantic action to be executed once the rule option has been recognised"

    def __init__(self,action):
        "Semantic Action Constructor"
        # Action to be executed
        self.action=action
        
    def execute(self,treeNode):
        "Behaviour execution. Takes 'nodes' as the tree nodes set"
        # Has the application been told to stop?
        if treeNode.application.isRunning:
            nodes=[treeNode]
            for child in treeNode.children:
                nodes.append(child)
            # Write is the new print function
            write=treeNode.application.window.writeText
            localContext={}
            localContext["nodes"]=nodes
            localContext["write"]=write
            localContext["nitrO"]=treeNode.application.nitrO
            # Code execution with its own contexts.
            # print '* To execute: '+self.action # NoComment=TraceParsing
            # *** MODIFIED FOR PFC
            try:
                exec(self.action,treeNode.application.applicationGlobalContext,localContext)
            except SemanticActionException, e:
                e.append(self.action)
                raise e
            except (str, Exception), e:
                raise SemanticActionException(e, self.action)
            # *** MODIFIED FOR PFC
            #print '* Finished: '+self.action # NoComment=TraceParsing
            #print 'Locals: ',localContext # NoComment=TraceParsing
            #print 'Globals: ',treeNode.application.applicationGlobalContext.keys() # NoComment=TraceParsing

# ************************  Reify Semantic Action ***************************

class ReifySemanticAction:
    "A semantic action to be executed once the rule option has been recognised"

    def __init__(self,action,nitrO):
        "Semantic Action Constructor"
        # Action to be executed
        self.action=action
        # Global nitrO object
        self.nitrO=nitrO
        
    def execute(self,treeNode):
        "Behaviour execution. Takes nitrO and application objects."
        # Has the application been told to stop?
        if treeNode.application.isRunning:
            # Write is the new print function
            write=treeNode.application.window.writeText
            localContext={}
            localContext["nitrO"]=self.nitrO
            localContext["write"]=write
            localContext["application"]=treeNode.application
            # Code execution with its own contexts.
            exec(self.action,treeNode.application.applicationGlobalContext,localContext)
    
    

# ******************** Module Testing Functions ********************

