"MetaLanguage Scanner and Parser Module"

import string
import langSpec

# ************************** Scanner ******************************
class Scanner:
    "MetaLanguage Definition Scanner Class"
    
    # Keywords
    LANGUAGE=256
    SCANNER=257
    PARSER=258
    SKIP=259
    NOTSKIP=260
    _REIFY_=261
    # Keyword List
    keywordList=[["LANGUAGE",LANGUAGE],
                 ["SCANNER",SCANNER],
                 ["PARSER",PARSER],
                 ["SKIP",SKIP],
                 ["_REIFY_",_REIFY_],
                 ["NOTSKIP",NOTSKIP]]
    # Tokens
    ID=270
    STRING=271
    ARROW=272
    OPENED_CURLY_BRACE=273
    CLOSED_CURLY_BRACE=274
    PIPE=275
    SEMICOLON=276
    EOF=277
    ASIGN=278
    CODE=279

    # String Exception
    stringException="Meta-Language Scanner Error"

    def __init__(self,string):
        "Constructor"
        # Last token
        self.token=0
        # Input String
        self.input=string
        # Buffer pointer
        self.charPos=0
        # Text Recognised
        self.text=""
        # Line number
        self.lineNumber=1

    def getChar(self):
        "Reads one Character"
        self.charPos=self.charPos+1
        char=self.input[self.charPos-1:self.charPos]
        if char=="\n":
            self.lineNumber=self.lineNumber+1
        return char
    
    def setChar(self):
        "Returns one Character to the Buffer"
        if self.input[self.charPos:self.charPos+1]=="\n":
            self.lineNumber=self.lineNumber-1
        self.charPos=self.charPos-1

    def error(self,msg=None):
        if msg:
            raise Scanner.stringException+" in line "+str(self.lineNumber)+"\n"+msg
        raise Scanner.stringException+" in line "+str(self.lineNumber)
    
    def getNextToken(self):
        "New token to be read"
        self.text=char=self.getChar()
        # End Of File
        if char=='':
            self.token=Scanner.EOF
        # Ignored Characters
        elif char==" " or char=="\n" or char=="\t":
            self.token=self.getNextToken()
        # Character |
        elif char=="|":
            self.token=Scanner.PIPE
        # Character ;
        elif char==";":
            self.token=Scanner.SEMICOLON
        # Character {
        elif char=="{":
            self.token=Scanner.OPENED_CURLY_BRACE
        # Character }
        elif char=="}":
            self.token=Scanner.CLOSED_CURLY_BRACE
        # Character =
        elif char=="=":
            self.token=Scanner.ASIGN
        # Token ->
        elif char=="-":
            char=self.getChar()
            if char==">":
                self.text="->"
                self.token=Scanner.ARROW
            else:
                 self.error("After the '-' char, the '>' is needed.")
        # Token STRING
        elif char=="\"":
            self.text=""
            char=self.getChar()
            while char!="\"":
                if char!="\"":
                    self.text=self.text+char
                    char=self.getChar()
            self.token=Scanner.STRING
        # Token ID
        elif char=="_" or (char>="a" and char<="z") or (char>="A" and char<="Z"):
            self.text=""
            while char=="_" or (char>="a" and char<="z") or (char>="A" and char<="Z") or (char>="0" and char<="9") :
                self.text=self.text+char
                char=self.getChar()
            self.setChar()
            self.token=self.idOrKeyword()
        # Token CODE
        elif char=="<":
            char=self.getChar()
            if char!="#":
                self.error("After the '<' char, the '#' is needed.")
            char=self.text=""
            codeLevel=1
            while codeLevel:
                char=self.getChar()
                if char=="#":
                    char=self.getChar()
                    if char!=">":
                        self.text=self.text+"#"+char
                    else:
                        codeLevel=codeLevel-1
                        if codeLevel:
                            self.text=self.text+"#>"
                elif char=="<":
                    self.text=self.text+char
                    char=self.getChar()
                    self.text=self.text+char
                    if char=="#":
                        codeLevel=codeLevel+1
                else:
                    self.text=self.text+char
            self.token=Scanner.CODE
        else:
            self.error("No token recognised.")
        return self.token

    def idOrKeyword(self):
        "Identifies weather the text is a keyword or an ID"
        text=string.upper(self.text)
        result=Scanner.ID
        for i in Scanner.keywordList:
            if i[0]==text:
                result=i[1]
                break
        return result
    
# ************************** Parser ******************************          
# Meta-Language LL1 Gramar
# <start> ->        LANGUAGE = ID <scan> <parser> <skip> <notskip>
# <scan> ->         SCANNER = { <scannerRules> }
# <parser> ->       PARSER = { <parserRules> }
# <skip> ->         SKIP = { <tokens> }
# <notskip> ->      NOTSKIP = { <tokens> }
# <scannerRules> ->
#                 | <SR> ; <moreSR>
# <moreSR> ->
#           |       <SR> ; <moreSR>
# <SR> ->           STRING ID -> <rightSR>
# <rightSR> ->      <rightSRItems> <ruleCode> <moreRightSR>
# <rightSRItems> ->
#               |   <rightSRItem> <moreRSRI>
# <rightSRItem> ->  STRING
#                 | ID
# <moreRSRI> ->     
#             |     <rightSRItem> <moreRSRI>
# <moreRightSR> ->
#           |       | <rightSRItems> <ruleCode> <moreRightSR>
# <parserRules> ->
#                 | <PR> ; <morePR>
# <morePR> ->
#           |       <PR> ; <morePR>
# <PR> ->           STRING ID -> <rightPR> <ruleCode> <moreRightPR>
# <rightPR> ->
#            |      ID <moreRPRI>
#            |      _REIFY_
# <ruleCode> -> 
#             |     CODE
# <moreRightPR> ->
#             |     | <rightPR> <ruleCode> <moreRightPR>
# <moreRPRI> ->
#             |     ID <moreRPRI>
# <tokens> ->
#           |       STRING ; <tokens>

class Parser:
    "MetaLanguage Definition Parser Class"
    
    # String Exception
    stringException="Meta-Language Parser Error"

    def error(self,msg=None):
        if msg:
            raise Parser.stringException+" in line "+str(self.scanner.lineNumber)+".\n"+msg
        raise Parser.stringException+" in line "+str(self.scanner.lineNumber)
    
    def __init__(self,text):
        "Parser Constructor"
        # Scanner Object
        self.scanner=Scanner(text)
        # Get the first token
        self.scanner.getNextToken()
        # Object-Grammar Creator
        self.ogc=ObjectGrammarCreator()
        
    def start(self):
        "<start> LL1 Free-Context Rule"
        # <start> -> LANGUAGE = ID <scan> <parser> <skip> <notskip>
        if self.scanner.token!=Scanner.LANGUAGE:
            self.error("First token must be LANGUAGE.")
        tok=self.scanner.getNextToken()
        if tok!=Scanner.ASIGN:
            self.error("= Token expected.")
        tok=self.scanner.getNextToken()
        if tok!=Scanner.ID:
            self.error("ID Token expected.")
        self.ogc.createLanguage(self.scanner.text)
        tok=self.scanner.getNextToken()
        self.scan()
        self.parser()
        self.skip()
        self.notskip()
        if self.scanner.token!=Scanner.EOF:
            self.error("End Of Program expected.")
            
    # <scan> -> SCANNER = { <scannerRules> }
    def scan(self):
        "<scanner> LL1 Free-Context Rule"
        if self.scanner.token!=Scanner.SCANNER:
            self.error("SCANNER Token expected.")
        tok=self.scanner.getNextToken()
        if tok!=Scanner.ASIGN:
            self.error("= Token expected.")
        tok=self.scanner.getNextToken()
        if tok!=Scanner.OPENED_CURLY_BRACE:
            self.error("{ Token expected.")
        tok=self.scanner.getNextToken()
        self.scannerRules()
        if self.scanner.token!=Scanner.CLOSED_CURLY_BRACE:
            self.error("} Token expected.")
        self.scanner.getNextToken()
            
    # <parser> -> PARSER = { <parserRules> }
    def parser(self):
        "<parser> LL1 Free-Context Rule"
        if self.scanner.token!=Scanner.PARSER:
            self.error("PARSER Token expected.")
        tok=self.scanner.getNextToken()
        if tok!=Scanner.ASIGN:
            self.error("= Token expected.")
        tok=self.scanner.getNextToken()
        if tok!=Scanner.OPENED_CURLY_BRACE:
            self.error("{ Token expected.")
        tok=self.scanner.getNextToken()
        self.parserRules()
        if self.scanner.token!=Scanner.CLOSED_CURLY_BRACE:
            self.error("} Token expected.")
        self.scanner.getNextToken()
        
        
    # <skip> -> SKIP = { <tokens> }
    def skip(self):
        "<skip> LL1 Free-Context Rule"
        if self.scanner.token!=Scanner.SKIP:
            self.error("SKIP Token expected.")
        self.ogc.enterSkipTokens()
        tok=self.scanner.getNextToken()
        if tok!=Scanner.ASIGN:
            self.error("= Token expected.")
        tok=self.scanner.getNextToken()
        if tok!=Scanner.OPENED_CURLY_BRACE:
            self.error("{ Token expected.")
        tok=self.scanner.getNextToken()
        self.tokens()
        if self.scanner.token!=Scanner.CLOSED_CURLY_BRACE:
            self.error("} Token expected.")
        self.scanner.getNextToken()
        
    # <notskip> -> NOTSKIP = { <tokens> }
    def notskip(self):
        "<notskip> LL1 Free-Context Rule"
        if self.scanner.token!=Scanner.NOTSKIP:
            self.error("NOTSKIP Token expected.")
        self.ogc.enterNotSkipTokens()
        tok=self.scanner.getNextToken()
        if tok!=Scanner.ASIGN:
            self.error("= Token expected.")
        tok=self.scanner.getNextToken()
        if tok!=Scanner.OPENED_CURLY_BRACE:
            self.error("{ Token expected.")
        tok=self.scanner.getNextToken()
        self.tokens()
        if self.scanner.token!=Scanner.CLOSED_CURLY_BRACE:
            self.error("} Token expected.")
        self.scanner.getNextToken()
        
    # <scannerRules> ->
    #                 | <SR> ; <moreSR>
    def scannerRules(self):
        "<scannerRules> LL1 Free-Context Rule"
        if self.scanner.token==Scanner.STRING:
            self.SR()
            tok=self.scanner.token
            if tok!=Scanner.SEMICOLON:
                self.error("; Token expected.")
            self.scanner.getNextToken()
            self.moreSR()
        
    # <moreSR> ->
    #           |       <SR> ; <moreSR>
    def moreSR(self):
        "<moreSR> LL1 Free-Context Rule"
        if self.scanner.token==Scanner.STRING:
            self.SR()
            tok=self.scanner.token
            if tok!=Scanner.SEMICOLON:
                self.error("; Token expected.")
            self.scanner.getNextToken()
            self.moreSR()
        
    # <SR> ->           STRING ID -> <rightSR>
    def SR(self):
        "<SR> LL1 Free-Context Rule"
        if self.scanner.token!=Scanner.STRING:
            self.error("STRING Token expected.")
        description=self.scanner.text
        tok=self.scanner.getNextToken()
        if tok!=Scanner.ID:
            self.error("ID Token expected.")
        self.ogc.createScannerRule(self.scanner.text,description)
        tok=self.scanner.getNextToken()
        if tok!=Scanner.ARROW:
            self.error("-> Token expected.")
        self.scanner.getNextToken()
        self.rightSR()


    # <rightSR> ->      <rightSRItems> <ruleCode> <moreRightSR>
    def rightSR(self):
        "<rightSR> LL1 Free-Context Rule"
        self.rightSRItems()
        self.ruleCode()
        self.moreRightSR()
        
    # <rightSRItems> ->
    #               |   <rightSRItem> <moreRSRI>
    def rightSRItems(self):
        "<rightSRItems> LL1 Free-Context Rule"
        tok=self.scanner.token
        if tok==Scanner.STRING or tok==Scanner.ID:
            self.rightSRItem()
            self.moreRSRI()
            
    # <rightSRItem> ->  STRING
    #                 | ID
    def rightSRItem(self):
        "<rightSRItem> LL1 Free-Context Rule"
        tok=self.scanner.token
        if tok==Scanner.STRING:
            self.ogc.addScannerRightTerminal(self.scanner.text)
        elif tok==Scanner.ID:
            self.ogc.addScannerRightNonTerminal(self.scanner.text)
        else:
            self.error("STRING or ID Token expected.")
        self.scanner.getNextToken()
        
    # <moreRSRI> ->     
    #             |     <rightSRItem> <moreRSRI>
    def moreRSRI(self):
        "<moreRSRI> LL1 Free-Context Rule"
        tok=self.scanner.token
        if tok==Scanner.STRING or tok==Scanner.ID:
            self.rightSRItem()
            self.moreRSRI()
        
    # <moreRightSR> ->  
    #           |       | <rightSRItems> <ruleCode> <moreRightSR>
    def moreRightSR(self):
        "<moreRightSR> LL1 Free-Context Rule"
        tok=self.scanner.token
        if tok==Scanner.PIPE:
            self.ogc.addScannerRightOption()
            self.scanner.getNextToken()
            self.rightSRItems()
            self.ruleCode()
            self.moreRightSR()

    # <parserRules> ->
    #                 | <PR> ; <morePR>
    def parserRules(self):
        "<parserRules> LL1 Free-Context Rule"
        if self.scanner.token==Scanner.STRING:
            self.PR()
            tok=self.scanner.token
            if tok!=Scanner.SEMICOLON:
                self.error("; Token expected.")
            self.scanner.getNextToken()
            self.morePR()
        
    # <morePR> ->
    #           |       <PR> ; <morePR>
    def morePR(self):
        "<morePR> LL1 Free-Context Rule"
        if self.scanner.token==Scanner.STRING:
            self.PR()
            tok=self.scanner.token
            if tok!=Scanner.SEMICOLON:
                self.error("; Token expected.")
            self.scanner.getNextToken()
            self.morePR()
        
    # <PR> ->           STRING ID -> <rightPR> <ruleCode> <moreRightPR>

    def PR(self):
        "<PR> LL1 Free-Context Rule"
        if self.scanner.token!=Scanner.STRING:
            self.error("STRING Token expected.")
        description=self.scanner.text
        tok=self.scanner.getNextToken()
        if tok!=Scanner.ID:
            self.error("ID Token expected.")
        self.ogc.createParserRule(self.scanner.text,description)
        tok=self.scanner.getNextToken()
        if tok!=Scanner.ARROW:
            self.error("-> Token expected.")
        tok=self.scanner.getNextToken()
        self.rightPR()
        self.ruleCode()
        self.moreRightPR()
        
    # <rightPR> ->
    #            |      ID <moreRPRI>
    #            |      _REIFY_
    def rightPR(self):
        "<rightPR> LL1 Free-Context Rule"
        if self.scanner.token==Scanner.ID:
            self.ogc.addParserRightNonTerminal(self.scanner.text)
            self.scanner.getNextToken()
            self.moreRPRI()
        if self.scanner.token==Scanner._REIFY_:
            self.scanner.getNextToken()
            self.ogc.addParserRightReify()
        
    # <ruleCode> -> 
    #             |     CODE
    def ruleCode(self):
        "<ruleCode> LL1 Free-Context Rule"
        if self.scanner.token==Scanner.CODE:
            self.ogc.addRuleCode(self.scanner.text)
            self.scanner.getNextToken()
        
    #<moreRightPR> ->
    #             |     | <rightPR> <ruleCode> <moreRightPR>
    def moreRightPR(self):
        "<moreRightPR> LL1 Free-Context Rule"
        if self.scanner.token==Scanner.PIPE:
            self.scanner.getNextToken()
            self.ogc.addParserRightOption()
            self.rightPR()
            self.ruleCode()
            self.moreRightPR()
        
    # <moreRPRI> ->
    #             |     ID <moreRPRI>
    def moreRPRI(self):
        "<moreRPRI> LL1 Free-Context Rule"
        if self.scanner.token==Scanner.ID:
            self.ogc.addParserRightNonTerminal(self.scanner.text)
            self.scanner.getNextToken()
            self.moreRPRI()
        
    # <tokens> ->
    #           |       STRING ; <tokens>
    def tokens(self):
        "<tokens> LL1 Free-Context Rule"
        if self.scanner.token==Scanner.STRING:
            self.ogc.addSkipOrNotSkipToken(self.scanner.text)
            tok=self.scanner.getNextToken()
            if tok!=Scanner.SEMICOLON:
                self.error("; Token expected.")
            self.scanner.getNextToken()
            self.tokens()
    
    
# ********************* ObjectGrammarCreator **************************          
class ObjectGrammarCreator:
    "Object Grammar Creator Class; takes a MetaLanguage Specification and creates its object tree definition"

    def __init__(self):
        "ObjectGrammarCreator"
        # Language object in which we create the object grammar (don't know)
        self.language=0
        # Last rule inserted
        self.lastRule=0
        # Tells whether the tokens inserted are from the Skip or NotSkip section
        self.skipTokensFlag=0
        # Tells if the rule inserted is the first one
        self.startSymbol=1
        
    def createLanguage(self,name):
        "Creates the Language object"
        self.language=langSpec.Language(name)
        
    def createScannerRule(self,name,description):
        "Creates a new scanner rule"
        # Finishing the past lastRule
        self.insertsLambdaIfNecessary()
        # Creates a new one
        self.lastRule=langSpec.NonTerminal(name,description)
        self.language.lexicalSpec[name]=self.lastRule
        
    def addScannerRightTerminal(self,text):
        "Adds a string on its last right option"
        lastOption=len(self.lastRule.options)-1
        self.lastRule.options[lastOption].nodes.append(langSpec.Terminal(text))
        
    def addScannerRightNonTerminal(self,name):
        "Adds a NonTerminal on its last right option"
        lastOption=len(self.lastRule.options)-1
        self.lastRule.options[lastOption].nodes.append(langSpec.NonTerminal(name))

    def addScannerRightOption(self):
        "Adds a new option to the right part of the last rule"
        self.lastRule.options.append(langSpec.Right())

    def insertsLambdaIfNecessary(self):
        "Inserts lamda symbol on the lastRule's right-part, if necessary"      
        # Is there any current rule?
        if self.lastRule!=0:
            # Any right part is empty?
            for n in self.lastRule.options:
                if len(n.nodes)==0:
                    # If so, lets write a Lambda object
                    n.nodes.append(langSpec.Lambda())

    def createParserRule(self,name,description):
        "Creates a new parser rule"
        # Finishing the past lastRule
        self.insertsLambdaIfNecessary()
        # Creates a new one
        self.lastRule=langSpec.NonTerminal(name,description)
        self.language.syntacticSpec[name]=self.lastRule
        # Is the starting symbol?
        if self.startSymbol:
            self.startSymbol=0
            self.language.start=self.lastRule

    def addParserRightNonTerminal(self,name):
        "Adds a NonTerminal on its last right option"
        lastOption=len(self.lastRule.options)-1
        self.lastRule.options[lastOption].nodes.append(langSpec.NonTerminal(name))

    def addParserRightReify(self):
        "Adds a Reify Symbol on its last right option"
        lastOption=len(self.lastRule.options)-1
        self.lastRule.options[lastOption].nodes.append(langSpec.Reify())

    def addParserRightOption(self):
        "Adds a new option to the right part of the last rule"
        self.lastRule.options.append(langSpec.Right())
        
    def addRuleCode(self,code):
        "Adds the the last rule behaviour (written in python)"
        lastOption=len(self.lastRule.options)-1
        self.lastRule.options[lastOption].actions.append(langSpec.SemanticAction(code))

    def enterSkipTokens(self):
        "The next Terminal symbols will be recognised as Skip tokens"
        # Has the last rule been finished?
        if self.lastRule!=0:
            self.insertsLambdaIfNecessary()
            self.lastRule=0
        self.skipTokensFlag=1

    def enterNotSkipTokens(self):
        "The next Terminal symbols will be recognised as NotSkip tokens"
        self.skipTokensFlag=0

    def addSkipOrNotSkipToken(self,token):
        "Adds a token; skipTokenFlag tells us its section"
        if self.skipTokensFlag:
            self.language.skipSpec.append(langSpec.Terminal(token))
        else:
            self.language.notSkipSpec.append(langSpec.Terminal(token))
           
# ******************** Module Testing Functions ********************

# Scanner Testing Function
def testScanner():
    "Scanner Class Testing"
    s="""   Language    =
                Python
            Scanner = {
                "Letter Token"
                letter -> "a" | "b" | "c" ;
                "Arrow Token"
                arrow -> "->" ;
                }
            Parser = {
                "Initial Free-Context Rule"
                S -> A B
                   ;
                "A rule"
                A -> letter
                   | arrow  <#
print "Python Code" # Python Comment
if a:
    print "More Python Code"
#>
                   ;
                "B rule"
                B -> A
                   |
                   ;
                }
            Skip={ " "; }
            NotSkip= { "\t"; "\n"; }
            """
    obj=Scanner(s)
    token=obj.getNextToken()
    while token!=Scanner.EOF:
        print "Token number:\t",obj.token,"\t",
        print "Token text:\t",obj.text,"\t",
        print "Line: ",obj.lineNumber
        token=obj.getNextToken()

# Parser Testing Function
def testParser():
    "Scanner Class Testing"
    s="""   Language    =
                Python
            Scanner = {
                "Letter Token"
                letter -> "a" | "b" | "c" ;
                "Arrow Token"
                arrow -> letter "->" ;
                }
            Parser = {
                "Initial Free-Context Rule"
                S -> A B
                   ;
                "A rule"
                A -> letter
                   | arrow  <#
print "Python Code" # Python Comment
if a:
    print "More Python Code"
#>
                   ;
                "B rule"
                B -> A
                   |
                   ;
                }
            Skip={ " "; }
            NotSkip = { "\t"; "\n"; }
            """
    obj=Parser(s)
    obj.start()
    print "Parsing OK"

# Application Description Function
def testAppDescribe():
    "Application describe() testing function"
    s="""   Language    =
                Python
            Scanner = {
                "Letter Token"
                letter -> "a" | "b" | "c" ;
                "Arrow Token"
                arrow -> letter "->" ;
                }
            Parser = {
                "Initial Free-Context Rule"
                S -> A B
                   ;
                "A rule"
                A -> letter
                   | arrow  <#
print "Python Code" # Python Comment
if a:
    print "More Python Code"
#>
                   ;
                "B rule"
                B -> A
                   |  <# 3 + 4 - a #>
                   ;
                }
            Skip={ "blank"; }
            NotSkip = { "tab"; "newline"; }
            """
    obj=Parser(s)
    obj.start()
    print obj.ogc.language.describe()

# Grammar Symbols Description Functions
def testGSDescribe():
    "Grammar Symbols describe() testing function"
    s="""   Language    =
                Python
            Scanner = {
                "Letter Token"
                letter -> "a" | "b" | "c" ;
                "Arrow Token"
                arrow -> letter "->" ;
                }
            Parser = {
                "Initial Free-Context Rule"
                S -> A B
                   ;
                "A rule"
                A -> letter
                   | arrow  <#
print "Python Code" # Python Comment
if a:
    print "More Python Code"
#>
                   ;
                "B rule"
                B -> A
                   |  <# 3 + 4 - a #>
                   ;
                }
            Skip={ "blank"; }
            NotSkip = { "tab"; "newline"; }
            """
    obj=Parser(s)
    obj.start()
    print "Lexical Specification Rules:"
    for key in obj.ogc.language.lexicalSpec.keys():
        print obj.ogc.language.lexicalSpec[key].describe()
    print "\nSyntactic Specification Rules:"
    for key in obj.ogc.language.syntacticSpec.keys():
        print obj.ogc.language.syntacticSpec[key].describe()
    print "\nSkip Specification Tokens:"
    for token in obj.ogc.language.skipSpec:
        print token.describe()
    print "\nNotSkip Specification Tokens:"
    for token in obj.ogc.language.notSkipSpec:
        print token.describe()
    print "\n\n\nRule Right-Part Grammar Symbols:"
    for key in obj.ogc.language.lexicalSpec.keys():
        rule=obj.ogc.language.lexicalSpec[key]
        for option in rule.options:
            for node in option.nodes:
                print node.describe()
    for key in obj.ogc.language.syntacticSpec.keys():
        rule=obj.ogc.language.syntacticSpec[key]
        for option in rule.options:
            for node in option.nodes:
                print node.describe()

