# -*- coding: ISO-8859-1 -*-

import pickle, weakref, objs
import anydbm
import bsddb
import threading

#=============================================================================

# La tabla de instancias

class InstanceTable:
	def __init__(self):
		self.instances = weakref.WeakValueDictionary()
		# Al aadirlas, la bsqueda de instancias nicas siempre
		# tendr xito, y como tienen otras referencias a ellas
		# nunca desaparecern del diccionario
		self.instances.update(objs.nonStorableInstances)

	def addInstance(self, inst):
		self.instances[inst.getID()] = inst

	def getInstance(self, id):
		return self.instances[id]

	def deleteInstance(self, inst):
		del self.instances[inst.getID()]

#=============================================================================

class Manager(object):
	def __init__(self, interpreter, application):
		self.interpreter = interpreter
		self.application = application
		self.instanceTable = InstanceTable()
		self.storages = {}
		self.policies = {}

		# El almacenamiento por defecto
		defaultStorage = SimpleStorage(interpreter, application.name)
		defaultStorageName = 'default'
		self.registerStorage(defaultStorageName, defaultStorage)
		self.storage = defaultStorage

		# La poltica por defecto
		defaultPolicy = SimplePolicy()
		defaultPolicyName = 'default'
		self.registerPolicy(defaultPolicyName, defaultPolicy)
		self.policy = defaultPolicy

	def getInstanceTable(self): return self.instanceTable

	def registerStorage(self, name, storage):
		if self.storages.has_key(name):
			raise ('Ya hay registrado un almacenamiento' \
				+ ' llamado %s') % name
		self.storages[name] = storage
		storage.setManager(self)
	def unregisterStorage(self, name):
		storage = getStorage(name)
		if self.storage is storage:
			raise 'El almacenamiento %s est en uso' % name
		del self.storages[name]
		return storage
	def getStorage(self, name):
		try:
			return self.storages[name]
		except KeyError, e:
			raise 'El almacenamiento %s no est definido' % name
	def setStorage(self, name):
		storage = self.getStorage(name)
		self.storage.commit()
		self.storage = storage

	def registerPolicy(self, name, policy):
		if self.policies.has_key(name):
			raise ('Ya hay registrada una poltica' \
				+ ' llamada %s') % name
		self.policies[name] = policy
		policy.setManager(self)
	def unregisterPolicy(self, name):
		policy = self.getPolicy(name)
		if self.policy is policy:
			raise 'La poltica %s est en uso' % name
		del self.policies[name]
		return policy
	def getPolicy(self, name):
		try:
			return self.policies[name]
		except KeyError, e:
			raise 'La poltica %s no est definida' % name
	def setPolicy(self, name):
		policy = self.getPolicy(name)
		self.policy.commit()
		self.policy = policy

	def retrieveObject(self, id):
		return self.storage.retrieveObject(id)

	def storeObject(self, instance):
		# Le pedimos al objeto que se guarde, as si sabe que no est
		# modificado podr abortar la operacin y ahorrarnos tiempo
		return instance.store(self.storage)

	def makePersistent(self, instance):
		instance.makePersistent(self)
		self.notifyModified(instance, True)

	def makeTransient(self, instance):
		instance.makeTransient()

	def commit(self):
		self.policy.commit()

	def commitStorage(self):
		self.storage.commit()

	def terminate(self):
		self.commit()
		for s in self.storages.values(): s.terminate()
		for p in self.policies.values(): p.terminate()

	def notifyModified(self, instance, modified):
		self.policy.notifyModified(instance, modified)

#=============================================================================

# Un medio de almacenamiento genrico + cach de objetos almacenados

class Storage(object):
	def __init__(self, interpreter):
		self.interpreter = interpreter

	def setManager(self, manager):
		self.manager = manager
		self.instanceTable = manager.getInstanceTable()

	def storeObject(self, instance):
		# Guardamos la instancia
		self.doStoreObject(instance.getID(), instance)

		# Aadimos la instancia a la tabla
		self.instanceTable.addInstance(instance)

	def retrieveObject(self, id):
		try:
			# Intentamos coger la instancia de la tabla
			instance = self.instanceTable.getInstance(id)
		except KeyError:
			# Recuperamos la instancia
			instance = self.doRetrieveObject(id)

			# Aadimos la instancia a la tabla
			# IMPORTANTE: hay que aadirla ANTES de restaurar sus
			# campos, o podemos tener referencias circulares que
			# causaran una recursividad infinita...
			self.instanceTable.addInstance(instance)

			# Restaura la clase y dems propiedades
			instance.restore(self.interpreter, self)

		return instance

#-----------------------------------------------------------------------------

class SimpleStorage(Storage):
	def __init__(self, interpreter, appName):
		Storage.__init__(self, interpreter)
		self.instances = {}
		self.appName = interpreter.getApplication().name
		fileName = '%s.%s.dat' % (appName, self.__class__.__name__)
		self.fileName = fileName
		try:
			self.load(fileName)
		except: # Deberamos filtrar excepciones...
			#print 'SimpleStorage: Error al cargar la tabla'
			pass

	def has_key(self, id):
		return self.instances.has_key(id)

	def doRetrieveObject(self, id):
		instRepr = self.instances[id]
		instance = pickle.loads(instRepr)
		return instance

	def doStoreObject(self, id, instance):
		instRepr = pickle.dumps(instance)
		self.instances[id] = instRepr

	def commit(self):
		self.dump(self.fileName)

	def terminate(self):
		pass

	def dump(self, filename):
		f = open(filename, 'wb')
		pickle.dump(self.instances, f)
		f.close()

	def load(self, filename):
		f = open(filename, 'rb')
		self.instances = pickle.load(f)
		f.close()

#-----------------------------------------------------------------------------

class DBMStorage(Storage):
	def __init__(self, interpreter, appName):
		Storage.__init__(self, interpreter)
		self.appName = interpreter.getApplication().name
		fileName = '%s.%s.dat' % (appName, self.__class__.__name__)
		self.fileName = fileName
		self.dbm = anydbm.open(fileName, 'c')

	def has_key(self, id):
		return self.dbm.has_key(id)

	def doRetrieveObject(self, id):
		instRepr = self.dbm[id]
		instance  = pickle.loads(instRepr)
		return instance

	def doStoreObject(self, id, instance):
		instRepr = pickle.dumps(instance)
		self.dbm[id] = instRepr

	def commit(self):
		self.dbm.sync()

	def terminate(self):
		self.dbm.close()

#-----------------------------------------------------------------------------

class BSDDBStorage(Storage):
	def __init__(self, interpreter, appName, format, *args):
		Storage.__init__(self, interpreter)
		funcs = {'hash': bsddb.hashopen,
			'bt': bsddb.btopen,
			'rn': bsddb.rnopen}
		try:
			openFunc = funcs[format]
		except KeyError, e:
			raise 'El formato "%s" no se reconoce' % format
		self.appName = interpreter.getApplication().name
		fileName = '%s.%s.%s.dat' \
			% (appName, self.__class__.__name__, format)
		self.fileName = fileName
		openArgs = (fileName, 'c') + tuple(args)
		self.db = openFunc(*openArgs)

	def has_key(self, id):
		return self.db.has_key(id)

	def doRetrieveObject(self, id):
		instRepr = self.db[id]
		instance  = pickle.loads(instRepr)
		return instance

	def doStoreObject(self, id, instance):
		instRepr = pickle.dumps(instance)
		self.db[id] = instRepr

	def commit(self):
		self.db.sync()

	def terminate(self):
		self.db.close()

#=============================================================================

# Poltica de almacenamiento

class StoragePolicy(object):
	def __init__(self):
		self.modifiedInstances = {}

	def setManager(self, manager):
		self.manager = manager

	def addInstance(self, instance):
		id = instance.getID()
		self.modifiedInstances[id] = instance

	def deleteInstance(self, instance):
		id = instance.getID()
		try:
			del self.modifiedInstances[id]
		except KeyError, e:
			pass

	def storeInstances(self):
		for inst in self.modifiedInstances.values():
			self.manager.storeObject(inst)
		self.manager.commitStorage()
		self.modifiedInstances = {}

#-----------------------------------------------------------------------------

class SimplePolicy(StoragePolicy):
	def __init__(self, maxChanges = 1):
		StoragePolicy.__init__(self)
		self.maxChanges = maxChanges
		self.numChanges = 0

	def notifyModified(self, instance, modified):
		if modified:
			self.addInstance(instance)
			self.numChanges += 1
			if self.numChanges == self.maxChanges:
				self.commit()

	def commit(self):
		self.numChanges = 0
		self.storeInstances()

	def terminate(self):
		pass

#-----------------------------------------------------------------------------

class TimedPolicy(StoragePolicy):
	def __init__(self, seconds):
		StoragePolicy.__init__(self)
		self.seconds = seconds
		self.timer = None
		self.lock = threading.Lock()

	def notifyModified(self, instance, modified):
		if modified:
			self.lock.acquire()
			self.addInstance(instance)
			if not self.timer:
					self.timer = threading.Timer(
						self.seconds, self.commit)
			self.lock.release()

	def commit(self):
		self.lock.acquire()
		if self.timer: self.timer.cancel()
		self.storeInstances()
		self.lock.release()

	def terminate(self):
		pass
